import requests
import re
from datetime import datetime
import json
from .resultclass import BroadbandResult

class VMChecker:
    def __init__(
        self,
        APIAddress="https://storeapi.virginmedia.com/api",
        storeAddress="https://store.virginmedia.com/serviceability"
    ):
        self.commonHeaders = {
            "X-Requested-With": "XMLHttpRequest",
        }
        self.APIAddress = APIAddress
        self.storeAddress = storeAddress


    def query_address(self, postCode, houseNumber, strictAddress=True):
        time_stamp = int(datetime.now().timestamp())
        session = requests.Session()

        # We need to get some tokens, this is made by an initial calls.
        # Subsequent API calls will succeed.
        # Then we can fetch the postcode to premise ID mapping from the API,
        # from this determine the premise ID to pass to the store.
        session.get(
            f'{self.APIAddress}/getToken?_={time_stamp}',
            headers=self.commonHeaders,
        )
        address_list = session.post(
            f'{self.APIAddress}/checkAddress',
            headers=self.commonHeaders,
            data={"postcode": postCode}
        )
        premise = self.determine_premise(address_list.text, houseNumber, strictAddress)

        # No premise match
        if len(premise) == 0:
            returnResult = BroadbandResult(
                data={
                    "addressData": False,
                    'serviceable': False,
                    'success': False,
                    'fttp': False,
                },
            )
            return returnResult

        # If premiseId is null, this property isn't servicable
        if not premise['premiseId']:
            returnResult = BroadbandResult(
                data={
                    "addressData": premise,
                    'serviceable': False,
                    'success': True,
                    'fttp': False,
                },
            )
            return returnResult

        # We do not know how to fetch the serviceability data from the public
        # API (if it is even possible). Instead we make a request to the store
        # webservice, passing the premise IDs from the API. We can then extract
        # the JSON object that is supposed to be passed down to the webclient.
        # This is not split into a seperate function as we still need the
        # session (I am not sure why).

        data = session.post(
            self.storeAddress,
            data={
                "siteId": premise['siteId'],
                "premiseId": premise['premiseId'],
            },
        )

        jsExtractRegex = 'var vmStoreJson= ([^;]+);'
        serviceability = re.search(jsExtractRegex, data.text).group(1)
        serviceability = json.loads(serviceability)
        serviceability['serviceability']['addressData'] = premise

        return VMResult(data=serviceability)


    # This is unreliable, I have seen like "34 B" for people who have 34B, we
    # should add some script to work this out, the same logic can be used for
    # cityfibre.
    @staticmethod
    def determine_premise(struct, number, strictAddress):
        data = json.loads(struct)
        number = f"{number}".lower()
        for premise in data['addressList']:
            if premise['displayAddress'].lower().startswith(f'{number} '):
                return premise
            elif premise['displayAddress'].startswith(f'{number}, '):
                return premise
        if strictAddress:
            return dict()

        data['addressList'][0]['looseMatch'] = True
        data['addressList'][0]['address'] = data['addressList'][0]['displayAddress']
        return data['addressList'][0]


class VMResult:
    def __init__(self, data):
        self._data = data['serviceability']
        self._data['success'] = True
        self._data['servicable'] = True
        self._data['serviceable'] = True

    @property
    def data(self):
        return self._data

    @property
    def json(self):
        return json.dumps(self._data)

    @property
    def fttp(self):
        delivery = self.data['networkServiceabilityDetails']['deliveryMethod']
        return delivery == "FTTP"
